# -*- coding: utf-8 -*-
# Description:
# OSP w3af_console Daemon wrapper
#
# Authors:
# Hani Benhabiles <hani.benhabiles@greenbone.net>
#
# Copyright:
# Copyright (C) 2014 Greenbone Networks GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

""" OSP w3af class. """

import os
import logging
from xml.dom.minidom import parse as xml_parse

from ospd.ospd import OSPDError, OSPDaemon
from ospd.misc import main as daemon_main
from ospd_w3af import __version__

import pexpect

logger = logging.getLogger(__name__)

OSPD_W3AF_DESCRIPTION = """
This scanner runs the 'w3af' scanner installed on the local system.

w3af is a Web Application Attack and Audit Framework. The project's goal is to
create a framework to help you secure web applications by finding and exploiting
all web application vulnerabilities.

For more information, see the w3af website:
    http://w3af.org/
"""

OSPD_W3AF_PARAMS = \
    {'profile':
     {'type': 'selection',
      'name': 'Scan profile',
      'default': 'fast_scan|audit_high_risk|full_audit|OWASP_TOP10|bruteforce|'
                 'empty_profile|web_infrastructure|full_audit_spider_man'
                 '|sitemap',
      'description': 'Scan profiles are predefined set of plugins and'
                     ' customized configurations.',},
     'w3af_timeout':
     {'type': 'integer',
      'name': 'w3af scan timeout',
      'default': 3600,
      'description': 'Time to wait for the w3af scan to finish.',},
     'target_port':
     {'type': 'integer',
      'name': 'Target port',
      'default': 80,
      'description': 'Port on target host to scan',},
     'use_https':
     {'type': 'boolean',
      'name': 'Use HTTPS',
      'default': 0,
      'description': 'Whether the target application is running over HTTPS',},
     'http_request_status':
     {'type': 'boolean',
      'name': 'Show HTTP request status',
      'default': 0,
      'description': 'Whether to show the HTTP request\'s status in results',},
     'http_request_headers':
     {'type': 'boolean',
      'name': 'Show HTTP request headers',
      'default': 0,
      'description': 'Whether to show the HTTP request\'s headers in results',},
     'http_response_status':
     {'type': 'boolean',
      'name': 'Show HTTP response status',
      'default': 0,
      'description': 'Whether to show the HTTP response\'s status in results',}}


def get_w3af_version():
    """ Finds w3af scanner's version is executable """
    output = pexpect.spawn('w3af_console --version')
    output.expect(pexpect.EOF)
    for line in output.before.split('\n'):
        if line.startswith("Version: "):
            return line.split()[1]
    return "Parsing error."

def create_w3af_script(scan_id, target, output_file, options):
    """ Returns path to a w3af script file for the scan_id scan. """

    profile = options.get('profile')
    logger.info("w3af scan using {0} profile.".format(profile))
    script_path = "/tmp/w3af-script-{0}-{1}".format(scan_id, target)
    port = options.get('port')
    with open(script_path, 'w') as file_path:
        file_path.write("profiles use {0}\n".format(profile))
        if options.get('use_https') == '0':
            target_url = 'http://{0}:{1}'.format(target, port)
        else:
            target_url = 'https://{0}:{1}'.format(target, port)
        file_path.write("target set target {0}\n".format(target_url))
        file_path.write("plugins\n")
        file_path.write("output xml_file\n")
        file_path.write("output config xml_file\n")
        file_path.write("set output_file {0}\n".format(output_file))
        file_path.write("back\n")
        file_path.write("back\n")
        file_path.write("start\n")
    return script_path

class OSPDw3af(OSPDaemon):

    """ Class for ospd-w3af daemon. """

    def __init__(self, certfile, keyfile, cafile):
        """ Initializes the ospd-w3af daemon's internal data. """
        super(OSPDw3af, self).__init__(certfile=certfile, keyfile=keyfile,
                                       cafile=cafile)
        self.server_version = __version__
        self.scanner_info['name'] = 'w3af'
        self.scanner_info['version'] = get_w3af_version()
        self.scanner_info['description'] = OSPD_W3AF_DESCRIPTION
        for name, param in OSPD_W3AF_PARAMS.items():
            self.add_scanner_param(name, param)

    def check(self):
        """ Checks that w3af_console is found and is executable. """
        try:
            output = pexpect.spawn('w3af_console')
            output.expect("w3af>>>")
        except pexpect.ExceptionPexpect as message:
            logger.error("Check for w3af_console failed: %s", message)
            return False
        return True

    def process_scan_params(self, params):
        """ params is directly from the XML """
        profiles_list = self.get_scanner_param_default('profile').split('|')
        for name, values in [('profile', profiles_list),
                             ('w3af_timeout', range(7200)),
                             ('target_port', range(65535)),
                             ('use_https', [0, 1]),
                             ('http_request_status', [0, 1]),
                             ('http_request_headers', [0, 1]),
                             ('http_response_status', [0, 1])]:
            if name not in params:
                params[name] = self.get_scanner_param_default(name)
            if params[name] not in [str(v) for v in values]:
                raise OSPDError('Invalid %s value' % name, 'start_scan')
        # Some needs to be renamed ...
        for source, dest in [('target_port', 'port'),
                             ('w3af_timeout', 'timeout'),
                             ('http_request_status', 'status'),
                             ('http_request_headers', 'headers'),
                             ('http_response_status', 'res_status')]:
            params[dest] = params[source]
            del params[source]
        return params

    def exec_scan(self, scan_id, target):
        """ Starts the w3af scanner for scan_id scan. """

        options = self.get_scan_options(scan_id)
        assert 'port' in options
        assert 'timeout' in options
        assert 'use_https' in options
        assert 'profile' in options
        assert 'status' in options
        assert 'headers' in options
        assert 'res_status' in options

        output_file = "/tmp/w3af-report-{0}-{1}".format(scan_id, target)
        script_file = create_w3af_script(scan_id, target, output_file, options)
        # Spawn process
        output = pexpect.spawn('w3af_console -s {0}'.format(script_file))
        output.timeout = int(options['timeout'])
        try:
            output.expect("Scan finished in ")
            output.expect("w3af>>>")
        except pexpect.TIMEOUT:
            logger.debug("w3af scan reached timeout.")
            self.handle_timeout(scan_id, target)
            if logger.getEffectiveLevel() > logging.DEBUG:
                os.remove(script_file)
            output.close(True)
            return
        try:
            output.close(True)
        except pexpect.ExceptionPexpect:
            logger.debug("Couldn't terminate pexpect session properly.")

        # Now, parse output_file and make multiple results
        self.store_scan_results(scan_id, target, output_file, options['status'],
                                options['headers'], options['res_status'])

        # Cleanup
        if logger.getEffectiveLevel() > logging.DEBUG:
            os.remove(script_file)
            os.remove(output_file)
        else:
            logger.debug("{0} not removed.".format(output_file))

    def store_scan_results(self, scan_id, target, output_file,
                           req_status=1, req_headers=1, res_status=1):
        """ Stores scan results from the XML output_file """

        xmldoc = xml_parse(output_file)
        # w3afrun/vulnerability => result_type.ALARM
        vulns = xmldoc.getElementsByTagName('vulnerability')
        for vuln in vulns:
            vuln_name = vuln.getAttribute('name')
            severity = vuln.getAttribute('severity').lower()
            if severity == 'information':
                vuln_sev = '0.0'
            if severity == 'low':
                vuln_sev = '2.5'
            elif severity == 'medium':
                vuln_sev = '5.0'
            elif severity == 'high':
                vuln_sev = '7.5'
            else:
                logger.debug("Unknown severity {0}.".format(severity))
                vuln_sev = ''
            desc_elem = vuln.getElementsByTagName('description')[0]
            vuln_desc = desc_elem.childNodes[0].nodeValue
            vuln_desc = vuln_desc.split("This vulnerability was found in ")[0]
            request = vuln.getElementsByTagName('httprequest')[0]
            if req_status:
                status = request.getElementsByTagName('status')[0]
                vuln_desc = ''.join([vuln_desc, '\n\nHTTP Request Status: ',
                                     status.childNodes[0].nodeValue, '\n'])
            if req_headers:
                headers = request.getElementsByTagName('header')
                vuln_desc = ''.join([vuln_desc, '\nHTTP Request Headers:'])
                h_str = str()
                for header in headers:
                    field = header.getAttribute('field')
                    content = header.getAttribute('content')
                    h_str = ''.join([h_str, '\n ', field, ': ', content])
                vuln_desc = ''.join([vuln_desc, h_str])
            response = vuln.getElementsByTagName('httpresponse')[0]
            if res_status:
                status = response.getElementsByTagName('status')[0]
                vuln_desc = ''.join([vuln_desc, '\n\nHTTP Response Status: ',
                                     status.childNodes[0].nodeValue, '\n'])

            self.add_scan_alarm(scan_id, host=target, name=vuln_name,
                                value=vuln_desc, severity=vuln_sev)
        # w3afrun/information => result_type.LOG
        information = xmldoc.getElementsByTagName('information')
        for info in information:
            info_name = info.getAttribute('name')
            desc_elem = info.getElementsByTagName('description')[0]
            info_desc = desc_elem.childNodes[0].nodeValue
            info_desc = info_desc.split("This vulnerability was found in ")[0]
            info_desc = info_desc.split("This information was found in ")[0]
            request = info.getElementsByTagName('httprequest')[0]
            if req_status:
                status = request.getElementsByTagName('status')[0]
                info_desc = ''.join([info_desc, '\n\nHTTP Request Status: ',
                                     status.childNodes[0].nodeValue, '\n'])
            if req_headers:
                headers = request.getElementsByTagName('header')
                info_desc = ''.join([info_desc, '\nHTTP Request Headers:'])
                h_str = str()
                for header in headers:
                    field = header.getAttribute('field')
                    content = header.getAttribute('content')
                    h_str = ''.join([h_str, '\n ', field, ': ', content])
                info_desc = ''.join([info_desc, h_str])
            response = info.getElementsByTagName('httpresponse')[0]
            if res_status:
                status = response.getElementsByTagName('status')[0]
                info_desc = ''.join([info_desc, '\n\nHTTP Response Status: ',
                                     status.childNodes[0].nodeValue, '\n'])

            self.add_scan_log(scan_id, host=target, name=info_name,
                              value=info_desc)
        # w3afrun/error => result_type.ERROR
        errors = xmldoc.getElementsByTagName('error')
        for error in errors:
            # Error text is directly within node, not within <description>
            self.add_scan_error(scan_id, host=target,
                                value=error.childNodes[0].nodeValue)


def main():
    """ OSP w3af main function. """
    daemon_main('OSPD - w3af_consle wrapper', OSPDw3af)
